const TAG = "listener";

let censored_cache = new Map();
let context_requests = [];
let do_cache = true;
let current_tab_id = null;
let sticker_collections = null;

/**********************************************************************************************************************
 * OPTIONS
 **********************************************************************************************************************/
browser.runtime.onMessage.addListener(handleMessage);

function handleMessage(request, sender, sendResponse) {
    if (request.getdata) {
        sendData(sendResponse);
        return true;
    }
    if (request.clearcache) {
        censored_cache = new Map();
    }
    if (request.reset_options) {
        restoreDefaultConfiguration();
    }
    if (request.initialize_options) {
        initializeOptions();
    }
    if (request.reset_stickers) {
        loadDefaultStickers();
    }
    if (request.options) {
        updateOptions(false);
    }
    if (request.logout) {
        logout();
    }
    if (request.login) {
        login(request.username, request.password, sendResponse);
        return true;
    }
}

function sendData(sendResponse) {
    sendResponse({
        cache: censored_cache,
        queue: queue
    });
}

let extension_active = true;
let labels = null;
let censor_type = null;
let file_types = null;


function initializeOptions() {
    loadDefaultStickers();
    let options_request = browser.storage.sync.get(['active', 'labels', 'censor_type', 'file_types', 'debug', 'do_cache',
        'bar_configuration', 'blur_configuration', 'pixel_configuration', 'glitch_configuration', 'triangle_configuration', 'sticker_configuration',
        'gif_configuration', 'png_configuration', 'jpg_configuration', 'bmp_configuration', 'webp_configuration',
        'whiteblacklist_configuration','reverse_mode_configuration', 'clustering_configuration', 'base64_scanner']);
    options_request.then((res) => {

        // Debugging
        if (res.debug != null) {
            debug = res.debug;
        } else {
            debug = false;
            browser.storage.sync.set({
                debug: false,
            });
        }

        //Cache
        if (res.do_cache != null) {
            do_cache = res.do_cache;
        } else {
            do_cache = true;
            browser.storage.sync.set({
                do_cache: true,
            });
        }

        // Active
        if (res.active != null) {
            extension_active = res.active;
        } else {
            extension_active = true;
            browser.storage.sync.set({
                active: true,
            });
        }
        // Labels
        if (res.labels != null) {
            labels = res.labels;
        } else {
        	// TODO obtain from list of areas
            labels = ["BELLYEXPOSED",
                "BELLYCOVERED",
                "BUTTOCKSEXPOSED",
                "BUTTOCKSCOVERED",
                "FEMALEBREASTEXPOSED",
                "FEMALEBREASTCOVERED",
                "FEMALEGENITALIAEXPOSED",
                "FEMALEGENITALIACOVERED",
                "MALEGENITALIACOVERED",
                "MALEGENITALIAEXPOSED",
                "MALEBREASTEXPOSED",
                "MALEBREASTCOVERED",
                "FEETCOVERED",
                "FEETEXPOSED",
                "ARMPITSCOVERED",
                "ARMPITSEXPOSED",
                "ANUSCOVERED",
                "ANUSEXPOSED"];
            browser.storage.sync.set({
                labels: labels
            });
        }
        // Censor Type
        if (res.censor_type != null) {
            censor_type = res.censor_type;
        } else {
            censor_type = 'black';
            browser.storage.sync.set({
                censor_type: 'black'
            });
        }
        // File Types
        if (res.file_types != null) {
            file_types = res.file_types;
        } else {
            file_types = ['jpg', 'png', 'bmp', 'webp'];
            browser.storage.sync.set({
                file_types: ['jpg', 'png', 'bmp', 'webp']
            });
        }

        // CensorType Configuration
        if (res.blur_configuration != null) {
            blur_configuration = res.blur_configuration;
        } else {
            browser.storage.sync.set({
                blur_configuration: default_blur_configuration,
            });
        }
        if (res.pixel_configuration != null) {
            pixel_configuration = res.pixel_configuration;
        } else {
            browser.storage.sync.set({
                pixel_configuration: default_pixel_configuration,
            });
        }
        if (res.bar_configuration != null) {
            bar_configuration = res.bar_configuration;
        } else {
            browser.storage.sync.set({
                bar_configuration: default_bar_configuration,
            });
        }

        if (res.glitch_configuration != null) {
            glitch_configuration = res.glitch_configuration;
        } else {
            browser.storage.sync.set({
                glitch_configuration: default_glitch_configuration,
            });
        }

        if (res.triangle_configuration != null) {
            triangle_configuration = res.triangle_configuration;
        } else {
            browser.storage.sync.set({
                triangle_configuration: default_triangle_configuration,
            });
        }

        if (res.sticker_configuration != null) {
            sticker_configuration = res.sticker_configuration
        } else {
            browser.storage.sync.set({
                sticker_configuration: default_sticker_configuration,
            });
        }

        if (res.png_configuration != null) {
            png_configuration = res.png_configuration;
        } else {
            browser.storage.sync.set({
                png_configuration: default_png_configuration,
            });
        }

        if (res.jpg_configuration != null) {
            jpg_configuration = res.jpg_configuration;
        } else {
            browser.storage.sync.set({
                jpg_configuration: default_jpg_configuration,
            });
        }

        if (res.bmp_configuration != null) {
            bmp_configuration = res.bmp_configuration;
        } else {
            browser.storage.sync.set({
                bmp_configuration: default_bmp_configuration,
            });
        }

        if (res.webp_configuration != null) {
            webp_configuration = res.webp_configuration;
        } else {
            browser.storage.sync.set({
                webp_configuration: default_webp_configuration,
            });
        }

        if (res.gif_configuration != null) {
            gif_configuration = res.gif_configuration;
        } else {
            browser.storage.sync.set({
                gif_configuration: default_gif_configuration,
            });
        }
        // Blacklist / Whitelist
        if (res.whiteblacklist_configuration != null) {
            whiteblacklist_configuration = res.whiteblacklist_configuration;
        } else {
            browser.storage.sync.set({
                whiteblacklist_configuration: default_whiteblacklist_configuration,
            });
        }

        if (res.reverse_mode_configuration != null) {
            reverse_mode_configuration = res.reverse_mode_configuration;
        } else {
            browser.storage.sync.set({
                reverse_mode_configuration: default_reverse_mode_configuration,
            });
        }

        // Clustering
        if (res.clustering_configuration != null) {
            clustering_configuration = res.clustering_configuration;
        } else {
            browser.storage.sync.set({
                clustering_configuration: default_clustering_configuration,
            });
        }

        if (res.base64_scanner != null) {
            // This option is not relevant for the background script
        } else {
            browser.storage.sync.set({
                base64_scanner: true,
            });
        }

        repairConfigurations();

        refreshLogin();
    });

    let local_request = browser.storage.local.get(['sticker_collections']);
    local_request.then((res) => {
        if (res.sticker_collections != null) {
            sticker_collections = res.sticker_collections;
        } else {
            browser.storage.local.set({
                sticker_collections: sticker_collections,
            });
        }
    });
}

function updateOptions(startup) {
    let options_request = browser.storage.sync.get(['active', 'labels', 'censor_type', 'file_types', 'bar_color', 'debug',
        'do_cache', 'bar_configuration', 'blur_configuration', 'pixel_configuration', 'glitch_configuration', 'triangle_configuration','sticker_configuration',
        'gif_configuration', 'png_configuration', 'jpg_configuration', 'bmp_configuration', 'webp_configuration',
        'whiteblacklist_configuration','reverse_mode_configuration','clustering_configuration']);
    options_request.then((res) => {

        debug = res.debug || false;
        if (extension_active !== res.active || startup) {
            browser.browsingData.remove({},
                {
                    cache: true,
                });
            log("Cache cleared");
        }

        extension_active = res.active;
        labels = res.labels || [];
        // Censor Type changed, unlock all locked
        if(censor_type != res.censor_type){
            for (let [url, cached_result] of censored_cache) {
                cached_result._lock_censortype = false;
            }
        }
        censor_type = res.censor_type || 'black';
        file_types = res.file_types || ['jpg', 'png', 'bmp', 'webp'];
        do_cache = res.do_cache;
        blur_configuration = res.blur_configuration || default_blur_configuration;
        bar_configuration = res.bar_configuration || default_bar_configuration;
        pixel_configuration = res.pixel_configuration || default_pixel_configuration;
        glitch_configuration = res.glitch_configuration || default_glitch_configuration;
        triangle_configuration = res.triangle_configuration || default_triangle_configuration;
        sticker_configuration = res.sticker_configuration || default_sticker_configuration;
        png_configuration = res.png_configuration || default_png_configuration;
        jpg_configuration = res.jpg_configuration || default_jpg_configuration;
        bmp_configuration = res.bmp_configuration || default_bmp_configuration;
        webp_configuration = res.webp_configuration || default_webp_configuration;
        gif_configuration = res.gif_configuration || default_gif_configuration;
        whiteblacklist_configuration = res.whiteblacklist_configuration || default_whiteblacklist_configuration;
        reverse_mode_configuration = res.reverse_mode_configuration || default_reverse_mode_configuration;
        clustering_configuration = res.clustering_configuration || default_clustering_configuration;
        if (reverse_mode_configuration.enabled == true && (user == null || user.permissions.permission_reverse_censoring > user.patreon_tier)) {
            reverse_mode_configuration.enabled = false;
            browser.storage.sync.set({
                reverse_mode_configuration: default_reverse_mode_configuration,
            });
        }
        let local_request = browser.storage.local.get(['sticker_collections']);
        local_request.then((res) => {
            if (res.sticker_collections != null) {
                sticker_collections = res.sticker_collections;
            } else {
                browser.storage.local.set({
                    sticker_collections: sticker_collections,
                });
            }
        });

    });
}

initializeOptions();


/**********************************************************************************************************************/


/**********************************************************************************************************************
 * AI
 **********************************************************************************************************************/
const [modelWeight, modelHeight] = [320, 320];
const weights = browser.runtime.getURL("web_model/model.json");
var model = null;
tf.loadGraphModel(weights).then(model_await => {
    model = model_await;
    //Warmup

    const dummy = tf.zeros([1, 320, 320, 3], 'float32');
    return model.executeAsync(dummy).then(function (result) {
        const [boxes, scores, classes, valid_detections] = result;
        valid_detections.dispose();
        classes.dispose();
        scores.dispose();
        boxes.dispose();
        dummy.dispose();
        log("Model LOADED");
        return result;
    });

});

browser.webRequest.onBeforeRequest.addListener(
    blockImages,
    {urls: ["<all_urls>"], types: ["image", "imageset", "media", "main_frame"]},
    ["blocking"]
);

browser.tabs.onRemoved.addListener(handleRemoved);

function handleRemoved(tabId, removeInfo) {
    log("Tab: " + tabId + " is closing");
    log("Window ID: " + removeInfo.windowId);
    log("Window is closing: " + removeInfo.isWindowClosing);
    log("queue length before: ", queue.length);
    queue.forEach(function (item, index, object) {
        if (item.tabId == tabId) {
            object.splice(index, 1);
            log('Removed from queue', object);
        }
    });
    log("Queue length after: ", queue.length);

}

browser.tabs.onUpdated.addListener(handleUpdated);

function handleUpdated(tabId, changeInfo, tabInfo) {
    //log("Updated tab: " + tabId);
    //log("Changed attributes: ");
    //log(changeInfo);
    //log("New tab Info: ");
    //log(tabInfo);
    if (tabInfo.active) {
        if (current_tab_id != tabInfo.id) {
            current_tab_id = tabInfo.id;
            log("current_tab_id", current_tab_id);
            queue.sort(compareCensorRequest);
        }
    }

    // What if we open a new web-page but there are ongion requests
    // This is bugged
    /*
    queue.forEach(function (item, index, object) {
      if(item.tabId == tabId && item.originUrl != tabInfo.url){
          object.splice(index, 1);
          log("Removing ",tabId, item.originUrl, tabInfo);
      }
    });
    */

}


browser.webRequest.onHeadersReceived.addListener(
    uncache,
    {urls: ["<all_urls>"], types: ["image", "imageset", "media", "main_frame"]}, //"main_frame"
    ["blocking", "responseHeaders"]
);

function uncache(e) {
    if (!extension_active) {
        return;
    }
    log("UNCACHING:", e);
    var asyncCacheControl = new Promise((resolve, reject) => {
        for (var i = e.responseHeaders.length - 1; i >= 0; --i) {
            if (e.responseHeaders[i].name == "Cache-Control") {
                e.responseHeaders.splice(i, 1);
            }
        }
        var cacheControl = {
            name: "Cache-Control",
            value: "no-cache, no-store, must-revalidate"
        };
        e.responseHeaders.push(cacheControl);
        resolve({responseHeaders: e.responseHeaders});
    });
    return asyncCacheControl;
}


function compareCensorRequest(a, b) {
    /*
    if(a.tabId != current_tab_id){
        return 1;
    }
    */
    if (a.tabId != b.tabId) {
        if (a.tabId == current_tab_id) {
            //log("Sortiere a nach vorne weil current ", a.tabId, "current = ", current_tab_id);
            return -1;
        }
        if (b.tabId == current_tab_id) {
            //log("Sortiere a nach vorne weil nicht current", a.tabId, "current = ", current_tab_id);
            return 1;
        }
    }
    if (a.img_dimension > b.img_dimension) {
        //log("Sortiere größeres Biild nach vorne");
        return -1;
    }
    if (a.img_dimension < b.img_dimension) {
        //log("Sortiere kleines Bild nach hinten");
        return 1;
    }
    return 0;
}

function blockImages(requestDetails) {
    if (!extension_active) {
        return {};
    }

    if(whiteblacklist_configuration.mode > 0){
        if(whiteblacklist_configuration.mode == 1){
            for (let regex of whiteblacklist_configuration.white_list) {
                let regExp = new RegExp(regex);
                if(regExp.test(requestDetails.url) || regExp.test(requestDetails.originUrl)){
                    return {};
                }
            }
        }else if(whiteblacklist_configuration.mode == 2){
            let blacklisted = false;
            for (let regex of whiteblacklist_configuration.black_list) {
                let regExp = new RegExp(regex);
                if(regExp.test(requestDetails.url) || regExp.test(requestDetails.originUrl)){
                    blacklisted = true;
                    break;
                }
            }
            if(!blacklisted){
                return {};
            }
        }
    }

    let str = requestDetails.url.substring(0, requestDetails.url.length - 4);
    let pathArray = requestDetails.url.split('/');
    let protocol = pathArray[0];
    let host = pathArray[2];
    let favicon_url = protocol + '//' + host + '/favicon';
    if (favicon_url === str) {
        return {};
    }

    let filter = browser.webRequest.filterResponseData(requestDetails.requestId);

    filter.onerror = event => {
        console.log(`Error: ${filter.error}`);
    }
    let stream = [];

    let streaming = false;

    filter.onstart = event => {
        requestContentType(requestDetails.url, function (content_type_header) {
            //When we encounter a video we do not want to wait until the filter has finished
            if (content_type_header != null && content_type_header.startsWith('video')) {
                streaming = true;
                return {};
            }
        }, function(e){
            console.log(e)
        });
    }

    filter.ondata = event => {
        if(streaming){
            for (let data of stream) {
                filter.write(data);
            }
            stream = [];
            filter.write(event.data);
        }else{
            stream.push(event.data);
        }
    }

    filter.onstop = event => {
        if (streaming) {
            filter.disconnect();
            return;
        }
        //let t0 = performance.now();
        let blob = new Blob(stream);
        let filename = requestDetails.url.split('/').pop();
        let file = new File([blob], filename);
        checkMime(file, function (content_type_header, ext, header, cont) {
            let crf = context_requests.filter(e => e.url === requestDetails.url);
            crf.forEach(f => context_requests.splice(context_requests.findIndex(e => e.name === f.name), 1));
            if (crf.length === 0 && !processableContentType(content_type_header, file_types, gif_configuration)) {
                writeOriginalFilestreamToFilter(filter, stream);
                return {};
            }
            if (do_cache && content_type_header !== "image/gif") {
                let cache_url = requestDetails.url;
                if (crf.length > 0) {
                    cache_url = crf[0].originalUrl;
                }
                let cached_result = censored_cache.get(cache_url);

                let censor_type_local = censor_type;
                if(cached_result && cached_result._lock_censortype){
                    censor_type_local = cached_result.censortype;
                }
                if (crf.length > 0) {
                    if(cached_result){
                        cached_result._lock_censortype = true;
                    }
                    censor_type_local = crf[0].censor_type;
                }
                if (cached_result) {
                    handleCache(cached_result, requestDetails, filter, censor_type_local);
                    return {};
                }
            }
            //let blob = new Blob(stream, {type: content_type_header});
            //let filename = requestDetails.url.split('/').pop();
            //let file = new File([blob], filename);
            let censor_type_processed = censor_type;
            if (crf.length > 0) {
                censor_type_processed = crf[0].censor_type;
            }

            if ((!gif_configuration._thumbnails || crf.length > 0) && content_type_header === "image/gif" &&
                (file_types.includes('gif') || crf.length > 0)
            ) {
                handleGif(requestDetails, blob, file, filter, stream, filename, content_type_header, ext, header , censor_type_processed);
            } else {
                if (crf.length === 0 && !cont) {
                    writeOriginalFilestreamToFilter(filter, stream);
                    return {};
                }
                handleImage(requestDetails, blob, file, filter, stream, filename, content_type_header, ext, header , censor_type_processed);
            }
        }, file_types, gif_configuration);
    }
}

function handleGif(requestDetails, blob, file, filter, stream, filename, content_type_header, ext, header , censor_type_processed, callback= null){
    if (user != null && user.permissions.permission_file_type_gif <= user.patreon_tier) {
        //For some reason gifs are loaded from ff cache and we have to reload them manually when triggered by context
        if(stream && stream.length === 0){
            toDataURL(requestDetails.url, function (imgage_data, content_type) {
                blob = new Blob([imgage_data], {type: content_type});
                    if (gif_configuration._memory_method > 0) {
                        processGIFMem(requestDetails, blob, file, filter, stream, filename, content_type_header, ext, header , censor_type_processed, callback);
                    } else {
                        processGIF(requestDetails, blob, file, filter, stream, filename, content_type_header, ext, header , censor_type_processed, callback);
                    }
            });
            return {};
        }
        let img_conf = gif_configuration;
        if(!checkImageFileSize(img_conf, file.size)){
            log("rejected by checkImageFileSize:", requestDetails, img_conf, 'image/gif', file.size);
            if(gif_configuration._thumbnail_fallback || file.size < gif_configuration._filesize_min){
                fallBackToThumbnail(requestDetails, blob, file, filter, stream, filename, content_type_header, ext, header , censor_type_processed, callback);
            }else{
                writeOriginalFilestreamToFilter(filter, stream);
            }
            return {};
        }
        if (gif_configuration._memory_method > 0) {
            processGIFMem(requestDetails, blob, file, filter, stream, filename, content_type_header, ext, header , censor_type_processed, callback);
        } else {
            processGIF(requestDetails, blob, file, filter, stream, filename, content_type_header, ext, header , censor_type_processed, callback);
        }
    } else {
        fallBackToThumbnail(requestDetails, blob, file, filter, stream, filename, content_type_header, ext, header , censor_type_processed, callback);
        return {};
    }
}

function fallBackToThumbnail(requestDetails, blob, file, filter, stream, filename, content_type_header, ext, header , censor_type_processed, callback = null){
    handleImage(requestDetails, blob, file, filter, stream, filename, content_type_header, ext, header , censor_type_processed, callback);
}

function handleImage(requestDetails, blob, file, filter, stream, filename, content_type_header, ext, header , censor_type_processed, callback = null){
    processImage(requestDetails, blob, file, filter, stream, filename, content_type_header, ext, header , censor_type_processed).then(r => {
        if(callback){
            callback(r);
        }
    }).catch(e => {
        if(stream != null && filter != null){
            writeOriginalFilestreamToFilter(filter, stream);
        }
        log('error', e)
        if(callback){
            callback(blob);
        }
        return null;
    });
}

function writeOriginalFilestreamToFilter(filter, stream){
    for (let data of stream) {
        filter.write(data);
    }
    filter.disconnect();
}

function handleCache(cached_result, requestDetails, filter, censor_type_local = null) {
    log("This request was loaded from cache: ", cached_result);
    if (
        cached_result.censortype == censor_type &&
        cached_result.labels.length === labels.length &&
        cached_result.labels.every(function (value, index) {
            return value === labels[index]
        })
    ) {
        //This is for future purpose
        cached_result.load_from_cache = true;
    } else {
        cached_result.load_from_cache = false;
    }
    //New options to apply:
    if (censor_type_local == null) {
            cached_result.censortype = censor_type;
    } else {
        cached_result.censortype = censor_type_local;
    }


    cached_result.labels = labels;
    applyFilter(requestDetails, cached_result, filter);
}


function processGIF(requestDetails, blob, file, filter, stream, filename, content_type_header, ext, header , censor_type_local = null, callback = null) {
    if (censor_type_local == null) {
        censor_type_local = censor_type;
    }
    let t0 = performance.now();
    blob.arrayBuffer().then(buffer => {
        let gif = window.gifuct.parseGIF(buffer);
        let frames = window.gifuct.decompressFrames(gif, true);
        if(gif_configuration._frame_count_max > 0 && frames.length > gif_configuration._frame_count_max){
            log("rejected by Frame count:", gif_configuration, 'image/gif', file.size);
            fallBackToThumbnail(requestDetails, blob, file, filter, stream, filename, content_type_header, ext, header , censor_type_local, callback);
            return;
        }
        let gif_encoder = new GIF({
            workers: 2,
            quality: 10,
            debug: debug,
            //background: '#fff',
            //transparent: 'rgba(0,0,0,255)',
            //transparent: 0,
        });
        let gifpromises = [];
        let last_frame = null;
        let transparency_set = false;
        let gif_base_width = null;
        let gif_base_height = null;
        let cache = {};
        frames.forEach(function (element, idx, array) {
            if (idx === 0) {
                gif_base_width = element.dims.width;
                gif_base_height = element.dims.height;
                gif_encoder.setOption('width', gif_base_width);
                gif_encoder.setOption('height', gif_base_height);
                log("GIF ENC Initialised");

                /* return a single frame
                var tempCanvas = document.createElement('canvas');
                tempCanvas.width = element.dims.width;
                tempCanvas.height = element.dims.height;
                var tempCtx = tempCanvas.getContext('2d');
                var frameImageData = tempCtx.createImageData(element.dims.width, element.dims.height);
                frameImageData.data.set(element.patch);
                tempCtx.putImageData(frameImageData, 0, 0);
                tempCanvas.toBlob(function(blob) {
                    var reader  = new FileReader();
                    reader.onloadend = function () {
                        resolve({redirectUrl: reader.result});
                    }
                    reader.readAsDataURL(blob);
                });
                return;
                */
            }
            if (element.disposalType > 1 && !transparency_set && 'transparentIndex' in element) {
                gif_encoder.setOption('transparent', 'rgba(0,0,0,${element.transparentIndex})');
                transparency_set = true;
            }

            //log("frames", element);
            let tempCanvas = document.createElement('canvas');
            let tempCtx = tempCanvas.getContext('2d');
            tempCanvas.width = gif_base_width;//element.dims.width;
            tempCanvas.height = gif_base_height;//element.dims.height;
            //tempCanvas.width = element.dims.width;
            //tempCanvas.height = element.dims.height;

            //var naturalHeight = element.dims.height;
            //var naturalWidth = element.dims.width;
            let frameImageData = tempCtx.createImageData(element.dims.width, element.dims.height);
            //let frameImageData = tempCtx.createImageData(gif_base_width, gif_base_height);

            //let dif_width = Math.max(gif_base_width-element.dims.width, 0);
            //let dif_height = Math.max(gif_base_height-element.dims.height, 0);

            if (element.disposalType <= 1 && 'transparentIndex' in element && last_frame) {
                //frameImageData.data.set(last_frame.patch);
                //frameImageData = last_frame;

                frameImageData.data.set(element.patch);

                tempCtx.putImageData(frameImageData, element.dims.left, element.dims.top, 0, 0, element.dims.width, element.dims.height);
                let correctedFrameImageData = tempCtx.getImageData(0, 0, gif_base_width, gif_base_height);

                frameImageData = mergeImageData([last_frame, correctedFrameImageData]);
                //tempCtx.putImageData(frameImageData, element.dims.left, element.dims.top,0,0, element.dims.width, element.dims.height);
                tempCtx.putImageData(frameImageData, 0, 0);
            } else {
                frameImageData.data.set(element.patch);
                tempCtx.putImageData(frameImageData, element.dims.left, element.dims.top, 0, 0, element.dims.width, element.dims.height);
            }
            last_frame = tempCtx.getImageData(0, 0, gif_base_width, gif_base_height);
            //const input = tf.image.resizeBilinear(tf.browser.fromPixels(tempCanvas), [modelWeight, modelHeight]).div(255.0).expandDims(0);
            const fromPixels = tf.browser.fromPixels(tempCanvas);
            const resized = tf.image.resizeBilinear(fromPixels, [modelWeight, modelHeight]);
            const divved = resized.div(255.0);
            const input = divved.expandDims(0);
            fromPixels.dispose();
            resized.dispose();
            divved.dispose();
            let promise = //new Promise(function(resolve, reject) {
                Queue.enqueue(() => new Promise((resolve, reject) => {
                    model.executeAsync(input).then(res => {
                        const [boxes, scores, classes, valid_detections] = res;
                        let valid_detections_sync = valid_detections.dataSync();
                        let classes_sync = classes.dataSync();
                        let boxes_sync = boxes.dataSync();
                        let scores_sync = scores.dataSync();
                        valid_detections.dispose();
                        classes.dispose();
                        scores.dispose();
                        boxes.dispose();
                        let marvinImage = new MarvinImage();
                        tempCanvas.toBlob(function (blob) {
                            let urlCreator = window.URL || window.webkitURL;
                            //Revoked
                            let marvinUrl = urlCreator.createObjectURL(blob);
                            marvinImage.load(marvinUrl, function () {
                                let processedRequest = new ProcessedCensorRequest(requestDetails.url, blob, gif_base_width, gif_base_height, 'image/gif',
                                    boxes_sync, scores_sync, classes_sync, valid_detections_sync,
                                    censor_type_local, labels);
                                let conf = bundleConfiguration();
                                paintCensorRequest(marvinImage, processedRequest, conf, cache);
                                tempCtx.putImageData(marvinImage.imageData, 0, 0);
                                URL.revokeObjectURL(marvinUrl);
                                resolve(
                                    {
                                        frame: idx,
                                        delay: element.delay ? element.delay : 100,
                                        disposalType: element.disposalType,
                                        processedData: tempCtx
                                    });

                            });

                        });
                    }).catch(function (err) {
                        log("gif error", err);
                        resolve(
                            {
                                frame: idx,
                                delay: element.delay,
                                disposalType: element.disposalType,
                                processedData: tempCtx
                            }
                        );
                    }).finally(fin => {
                        input.dispose();
                        log('Tensorcount:', tf.memory().numTensors);
                    });

                }), requestDetails.tabId, requestDetails.originUrl, requestDetails.url, blob, gif_base_width, gif_base_height);
            gifpromises.push(promise);
        });

        Promise.allSettled(gifpromises).then((results) => results.forEach(
            (result) => {
                gif_encoder.addFrame(result.value.processedData, {
                    delay: result.value.delay,
                    dispose: result.value.disposalType
                });
            }
        )).finally(function (err) {
            gif_encoder.on('finished', function (blob) {
                if(filter){
                    blob.arrayBuffer().then(buffer => {
                        let t1 = performance.now();
                        log("Processing GIF time " + (t1 - t0) + " millisseconds.");
                        try {
                            filter.write(buffer);
                            filter.close();
                        } catch (err) {
                            log('Caught ' + err, buffer);
                        }
                        if(callback){
                            callback(blob);
                        }
                    });
                }else{
                    if(callback){
                        callback(blob);
                    }
                }
            });
            gif_encoder.render();

        });
    });
}

function processGIFMem(requestDetails, blob, file, filter, stream, filename, content_type_header, ext, header , censor_type_local = null, callback = null) {
    Queue.enqueue(() => new Promise((resolve, reject) => {
            if (censor_type_local == null) {
                censor_type_local = censor_type;
            }
            let t0 = performance.now();
            blob.arrayBuffer().then(buffer => {
                let gif = window.gifuct.parseGIF(buffer);
                let frames = window.gifuct.decompressFrames(gif, true);
                if(gif_configuration._frame_count_max > 0 && frames.length > gif_configuration._frame_count_max){
                    log("rejected by Frame count:", gif_configuration, 'image/gif');
                    fallBackToThumbnail(requestDetails, blob, file, filter, stream, filename, content_type_header, ext, header , censor_type_local, callback);
                    reject("rejected by Frame count");
                    return;
                }
                let gif_encoder = new GIF({
                    workers: 2,
                    quality: 10,
                    debug: debug,
                    //background: '#fff',
                    //transparent: 'rgba(0,0,0,255)',
                    //transparent: 0,
                });
                let gifpromises = [];
                let last_frame = null;
                let transparency_set = false;
                let gif_base_width = null;
                let gif_base_height = null;
                let cache = {};
                frames.forEach(function (element, idx, array) {
                    if (idx === 0) {
                        gif_base_width = element.dims.width;
                        gif_base_height = element.dims.height;
                        gif_encoder.setOption('width', gif_base_width);
                        gif_encoder.setOption('height', gif_base_height);
                        log("GIF ENC Initialised");

                        /* return a single frame
                        var tempCanvas = document.createElement('canvas');
                        tempCanvas.width = element.dims.width;
                        tempCanvas.height = element.dims.height;
                        var tempCtx = tempCanvas.getContext('2d');
                        var frameImageData = tempCtx.createImageData(element.dims.width, element.dims.height);
                        frameImageData.data.set(element.patch);
                        tempCtx.putImageData(frameImageData, 0, 0);
                        tempCanvas.toBlob(function(blob) {
                            var reader  = new FileReader();
                            reader.onloadend = function () {
                                resolve({redirectUrl: reader.result});
                            }
                            reader.readAsDataURL(blob);
                        });
                        return;
                        */
                    }
                    if (element.disposalType > 1 && !transparency_set && 'transparentIndex' in element) {
                        gif_encoder.setOption('transparent', 'rgba(0,0,0,${element.transparentIndex})');
                        transparency_set = true;
                    }
                    let promise = new Promise(function (resolve, reject) {
                        let tempCanvas = document.createElement('canvas');
                        let tempCtx = tempCanvas.getContext('2d');
                        tempCanvas.width = gif_base_width;//element.dims.width;
                        tempCanvas.height = gif_base_height;//element.dims.height;
                        let frameImageData = tempCtx.createImageData(element.dims.width, element.dims.height);
                        if (element.disposalType <= 1 && 'transparentIndex' in element && last_frame) {
                            frameImageData.data.set(element.patch);
                            tempCtx.putImageData(frameImageData, element.dims.left, element.dims.top, 0, 0, element.dims.width, element.dims.height);
                            let correctedFrameImageData = tempCtx.getImageData(0, 0, gif_base_width, gif_base_height);
                            frameImageData = mergeImageData([last_frame, correctedFrameImageData]);
                            tempCtx.putImageData(frameImageData, 0, 0);
                        } else {
                            frameImageData.data.set(element.patch);
                            tempCtx.putImageData(frameImageData, element.dims.left, element.dims.top, 0, 0, element.dims.width, element.dims.height);
                        }
                        last_frame = tempCtx.getImageData(0, 0, gif_base_width, gif_base_height);
                        const fromPixels = tf.browser.fromPixels(tempCanvas);
                        const resized = tf.image.resizeBilinear(fromPixels, [modelWeight, modelHeight]);
                        const divved = resized.div(255.0);
                        const input = divved.expandDims(0);
                        fromPixels.dispose();
                        resized.dispose();
                        divved.dispose();
                        log('Tensorcount:', tf.memory().numTensors);
                        model.executeAsync(input).then(res => {
                            input.dispose();
                            const [boxes, scores, classes, valid_detections] = res;
                            let valid_detections_sync = valid_detections.dataSync();
                            let classes_sync = classes.dataSync();
                            let boxes_sync = boxes.dataSync();
                            let scores_sync = scores.dataSync();
                            valid_detections.dispose();
                            classes.dispose();
                            scores.dispose();
                            boxes.dispose();
                            let marvinImage = new MarvinImage();
                            tempCanvas.toBlob(function (blob) {
                                let urlCreator = window.URL || window.webkitURL;
                                //Revoked
                                let marvinUrl = urlCreator.createObjectURL(blob);
                                marvinImage.load(marvinUrl, function () {
                                    let processedRequest = new ProcessedCensorRequest(requestDetails.url, blob, gif_base_width, gif_base_height, 'image/gif',
                                        boxes_sync, scores_sync, classes_sync, valid_detections_sync,
                                        censor_type_local, labels);
                                    let conf = bundleConfiguration();

                                    paintCensorRequest(marvinImage, processedRequest, conf, cache);
                                    tempCtx.putImageData(marvinImage.imageData, 0, 0);
                                    URL.revokeObjectURL(marvinUrl);
                                    resolve(
                                        {
                                            frame: idx,
                                            delay: element.delay ? element.delay : 100,
                                            disposalType: element.disposalType,
                                            processedData: tempCtx
                                        });
                                });
                            });
                        }).catch(function (err) {
                            log("gif error", err);
                            resolve(
                                {
                                    frame: idx,
                                    delay: element.delay,
                                    disposalType: element.disposalType,
                                    processedData: tempCtx
                                }
                            );
                        });
                    });
                    gifpromises.push(promise);
                });
                Promise.allSettled(gifpromises).then((results) => results.forEach(
                    (result) => {
                        gif_encoder.addFrame(result.value.processedData, {
                            delay: result.value.delay,
                            dispose: result.value.disposalType
                        });
                    }
                )).finally(function (err) {
                    gif_encoder.on('finished', function (blob) {
                        if(filter) {
                            blob.arrayBuffer().then(buffer => {
                                let t1 = performance.now();
                                log("Processing time " + (t1 - t0) + " millisseconds.");
                                try {
                                    filter.write(buffer);
                                    filter.close();
                                    resolve();
                                } catch (err) {
                                    log('Caught ' + err, buffer);
                                }
                                if(callback){
                                    callback(blob);
                                }
                            });
                        }else{
                            if(callback){
                                callback(blob);
                            }
                        }
                    });
                    gif_encoder.render();

                });
            });
        }), requestDetails.tabId, requestDetails.originUrl, requestDetails.url, blob, 9999
    );
}

function mergeImageData(imageDataArray) {

    let newImageData = imageDataArray[0];

    for (let j = 0; j < imageDataArray.length; j++) { // iterate through the imageDataArray

        for (let i = 0, bytes = imageDataArray[j].data.length; i < bytes; i += 4) { // iterate through image bytes

            let index = (imageDataArray[j].data[i + 3] === 0 ? 0 : j);

            newImageData.data[i] = imageDataArray[index].data[i];
            newImageData.data[i + 1] = imageDataArray[index].data[i + 1];
            newImageData.data[i + 2] = imageDataArray[index].data[i + 2];
            newImageData.data[i + 3] = imageDataArray[index].data[i + 3];
        }
    }
    return newImageData;
};

function processImage(requestDetails, blob, file, filter, stream, filename, content_type, ext, header, censor_type_local = null) {
    return new Promise((resolve, reject) => {
        if (!filename.endsWith("." + ext)) {
            filename += "." + ext;
            file = new File([blob], filename);
        }
        let img_conf = getImageConfiguration(content_type);
        if(content_type != 'image/gif' && !checkImageFileSize(img_conf, file.size)){
            log("rejected by checkImageFileSize:", requestDetails, img_conf, content_type, file.size);
            reject("rejected by checkImageFileSize");
            return;
        }
        if (censor_type_local == null) {
            censor_type_local = censor_type;
        }
        let urlCreator = window.URL || window.webkitURL;
        //Revoked
        let imageUrl = urlCreator.createObjectURL(blob);
        //let c = document.createElement('img');
        //c.src = imageUrl;
        let imageData = new Image();
        imageData.onload = function () {
            let naturalWidth = this.naturalWidth;
            let naturalHeight = this.naturalHeight;
            if(content_type != 'image/gif' && !checkImageDimensions(img_conf, this.naturalWidth, this.naturalHeight)){
                log("rejected by checkImageDimensions:", requestDetails, img_conf, content_type, this.naturalWidth, this.naturalHeight);
                reject("rejected by checkImageDimensions");
                return;
            }
            let img_dimension = this.naturalWidth * this.naturalHeight;
            Queue.enqueue(
                () => new Promise((resolve, reject) => {
                    //const input = tf.image.resizeBilinear(tf.browser.fromPixels(c), [modelWeight, modelHeight]).div(255.0).expandDims(0);
                    const fromPixels = tf.browser.fromPixels(imageData);
                    const resized = tf.image.resizeBilinear(fromPixels, [modelWeight, modelHeight]);
                    const divved = resized.div(255.0);
                    const input = divved.expandDims(0);
                    let t0 = performance.now();
                    model.executeAsync(input).then(res => {
                        let t1 = performance.now();
                        //console.log("model.executeAsync time " + (t1 - t0) + " millisseconds.");
                        const [boxes, scores, classes, valid_detections] = res;

                        let valid_detections_sync = valid_detections.dataSync();
                        let classes_sync = classes.dataSync();
                        let boxes_sync = boxes.dataSync();
                        let scores_sync = scores.dataSync();
                        valid_detections.dispose();
                        classes.dispose();
                        scores.dispose();
                        boxes.dispose();
                        log(valid_detections_sync, classes_sync, boxes_sync, scores_sync);
                        var processedRequest = new ProcessedCensorRequest(requestDetails.url, file, naturalWidth, naturalHeight, content_type,
                            boxes_sync, scores_sync, classes_sync, valid_detections_sync,
                            censor_type_local, labels);
                        if (do_cache && !censored_cache.get(requestDetails.url)) {
                            censored_cache.set(requestDetails.url, processedRequest);
                        }
                        applyFilter(requestDetails, processedRequest, filter, img_conf)
                            .catch(e => {
                                reject('applyFilter filter error');
                                return;
                            }).then(blob => {
                                resolve(blob);
                            });

                    }).catch(e => {
                        reject(e.message);
                        return;
                    }).finally(fin => {
                        resized.dispose();
                        divved.dispose();
                        input.dispose();
                        fromPixels.dispose();
                        URL.revokeObjectURL(imageUrl);
                        return;
                    });
                    log('Tensorcount:', tf.memory().numTensors);
                }).catch(e => {
                    URL.revokeObjectURL(imageUrl);
                    console.log('Tensorflow-Error', e, requestDetails.url);
                    reject('Tensorflow-Error');
                }).then((val) =>
                {
                    resolve(val)
                }
                )
                , requestDetails.tabId, requestDetails.originUrl, requestDetails.url, imageData, img_dimension)
            queue.sort(compareCensorRequest);
        };
        imageData.src = imageUrl;

    });
}

function applyFilter(requestDetails, processedRequest, filter, img_conf = null) {
    let promise = new Promise((resolve, reject) => {
        let imgdata = processedRequest.imgdata;
        let content_type = processedRequest.content_type;
        let marvinImage = new MarvinImage();
        let urlCreator = window.URL || window.webkitURL;
        let blob = new Blob([imgdata], {type: content_type});
        //Revoked
        let marvinUrl = urlCreator.createObjectURL(blob);
        marvinImage.load(marvinUrl, function () {
            //let t0 = performance.now();
            let conf = bundleConfiguration();
            let cache = {};
            paintCensorRequest(marvinImage, processedRequest, conf, cache);
            //let t1 = performance.now();
            //console.log("paintCensorReques time " + (t1 - t0) + " millisseconds.");
            if(img_conf === null){
                img_conf = getImageConfiguration(content_type);
            }
            let marvinBlob = marvinImage.toBlob(img_conf._file_output_type);
            if(filter) {
                marvinBlob.arrayBuffer().then(buffer => {
                    try {
                        filter.write(buffer);
                        filter.disconnect();
                        URL.revokeObjectURL(marvinUrl);
                        resolve(marvinBlob);
                    } catch (err) {
                        log('Caught ' + err, buffer);
                        URL.revokeObjectURL(marvinUrl);
                        reject();
                    }
                });
            }else{
                URL.revokeObjectURL(marvinUrl);
                resolve(marvinBlob);
            }
        });
    });
    return promise;
}

function getImageConfiguration(content_type, filesize, width, height){
    let conf = png_configuration;
    switch (content_type) {
        case 'image/png':
            conf = png_configuration;
            break;
        case 'image/jpg':
            conf = jpg_configuration;
            break;
        case 'image/jpeg':
            conf = jpg_configuration;
            break;
        case 'image/bmp':
            conf = bmp_configuration;
            break;
        case 'image/webp':
            conf = webp_configuration;
            break;
        case 'image/gif':
            conf = gif_configuration;
            break;
    }
    return conf;
}
function checkImageDimensions(conf, width, height){
    if( width < conf._width_min ||
        (conf._width_max > 0 && width > conf._width_max) ||
        height < conf._height_min ||
        (conf._height_max > 0 && height > conf._height_max)){
        return false;
    }
    return true;
}
function checkImageFileSize(conf, filesize){
    if( filesize < conf._filesize_min ||
        (conf._filesize_max > 0 && filesize > conf._filesize_max)
        ){
        return false;
    }
    return true;
}
